/******************************************************************************
 * ClientConnection.java - Manage a connection to the FIBS server
 * $Id: ClientConnection.java,v 1.3 2010/03/03 13:12:21 inim Exp $
 * 
 * BuckoFIBS - Backgammon by BuckoSoft
 * Copyright© 2009,2010 - Dick Balaska - BuckoSoft, Corp.
 * 
 * $Log: ClientConnection.java,v $
 * Revision 1.3  2010/03/03 13:12:21  inim
 * Replaced (c) sign in comment mangled by CVS default encoding back to UTF-8
 *
 * Revision 1.2  2010/03/03 12:19:49  inim
 * Moved source to UTF8 encoding from CP1252 encoding. To this end all source files' (c) message was updated to "Copyright© 2009,2010 - Dick Balaska - BuckoSoft, Corp.". This replaces the (c) sign to UTF8, and adds the new year 2010.
 *
 * Revision 1.1  2010/02/04 05:57:53  inim
 * Mavenized project folder layout
 *
 * Revision 1.20  2009/02/27 05:49:50  dick
 * Check for bare cr and bare lf at the beginning of a line.
 *
 * Revision 1.19  2009/02/25 08:12:14  dick
 * Check for a leading line feed, not a cr.
 *
 * Revision 1.18  2009/02/24 05:36:59  dick
 * Handle connectionAborted (on write).
 *
 * Revision 1.17  2009/02/17 14:45:44  dick
 * Wrap debug with DEBUG.
 *
 * Revision 1.16  2009/02/14 12:29:19  dick
 * Check for runon messages with just a cr between them.
 *
 * Revision 1.15  2009/02/02 08:38:12  dick
 * login with the app signature instead of hardcoded BuckoFIBS.
 *
 * Revision 1.14  2009/01/29 08:27:42  dick
 * Fix the pushback handling.
 *
 * Revision 1.13  2009/01/28 19:42:28  dick
 * Prettier cvs link in the javadoc.
 *
 * Revision 1.12  2009/01/28 08:30:24  dick
 * Handle the Malformed Messages (Bad_board) as documented as occuring in http://www.fibs.com/fcm/ .
 * However, we do begin and end regexp on ALL messages.
 * ClientConnection now supports a pushback buffer so that CookieMonster can deal with the second message later.
 *
 * Revision 1.11  2009/01/27 19:17:33  dick
 * Display the cookie number with the string in stdout.
 *
 * Revision 1.10  2009/01/26 17:35:15  dick
 * Push the CookieMonster down to the ClientConnection.  ClientAdapter now emits the cookie with the string.
 * There are many messages besides the known "BadBoard" messages that are a result of runon messages (missing crlf).
 * CookieMonster is going to need to do a full regexp scan and see if there is anything on the right that needs to be pushed back as another message.
 *
 * Revision 1.9  2009/01/12 21:48:53  dick
 * Add setFibsAttributes.
 *
 * Revision 1.8  2009/01/12 07:47:07  dick
 * Update to use the interfaces as ClientConnection is used across multiple applications.
 *
 * Revision 1.7  2009/01/09 07:37:48  dick
 * Turn off debug.
 *
 * Revision 1.6  2009/01/08 18:33:49  dick
 * Only shutdown the socket if it exists.
 *
 * Revision 1.5  2009/01/07 04:29:16  dick
 * Javadoc.
 *
 * Revision 1.4  2009/01/07 02:36:56  dick
 * Better handling when destroying ourselves.
 *
 * Revision 1.3  2009/01/05 07:17:20  dick
 * Use the selected profileId when determining values for the connection.
 *
 * Revision 1.2  2008/12/17 04:39:19  dick
 * Close the socket before deleting it.
 *
 * Revision 1.1  2008/12/13 06:57:11  dick
 * Move ClientConnection, CookieMonster, and FIBSMessages to their own network package.
 *
 * Revision 1.8  2008/12/13 06:19:45  dick
 * getDisplayXmit() becomes isDisplayXmit().
 *
 * Revision 1.7  2008/12/12 14:44:41  dick
 * Handle partial lines received.
 *
 * Revision 1.6  2008/12/11 20:25:01  dick
 * Echo the Xmit messages to the screen if pref enabled.
 *
 * Revision 1.5  2008/12/10 17:58:12  dick
 * Fix the list lock.
 * Output network messages in color.
 *
 * Revision 1.4  2008/12/09 19:31:44  dick
 * close the sockets on shutdown.
 *
 * Revision 1.3  2008/12/09 01:48:24  dick
 * Buffer receiving doesn't monitor for logins.  CookieMonster does that now.
 *
 * Revision 1.2  2008/03/31 07:07:41  dick
 * Login using clips message handling.
 *
 * Revision 1.1  2008/03/30 05:40:33  dick
 * Early BuckoFIBS networking.  We connect.
 *
 */

/* 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * The Original Code is BuckoFIBS, <http://www.buckosoft.com/BuckoFIBS/>.
 * The Initial Developer of the Original Code is Dick Balaska and BuckoSoft, Corp.
 * 
 */
package com.buckosoft.fibs.net;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.LinkedList;


/** Manage a connection to the FIBS server.
 * This object runs in it's own Thread.
 * The only synchronized object is the outbound message queue.
 * @author Dick Balaska
 * @since 2008/03/29
 * @version $Revision: 1.3 $ <br> $Date: 2010/03/03 13:12:21 $
 * @see <a href="http://cvs.buckosoft.com/Projects/BuckoFIBS/BuckoFIBS/src/com/buckosoft/fibs/net/ClientConnection.java">cvs ClientConnection.java</a>
 */
public class ClientConnection extends Thread {
	private	final static boolean DEBUG = false;
	
	final static String	eol = "\r\n";
	
	private	Socket	sock = null;
	private	InputStream is = null;
	private	OutputStream os = null;
	boolean	loggedIn = false;

	private	ClientAdapter		clientAdapter = null;
	private	CookieMonster		cookieMonster;
	private	FIBSAttributes		fibsAttributes;
	private	boolean				shuttingDown = false;
	private	LinkedList<String>	outMessages = new LinkedList<String>();
	private	Boolean				listLock = new Boolean(false);
	private	String				leftover = null;
	private	String				pushbackString = null;

	public ClientConnection() {
		cookieMonster = new CookieMonster();
		cookieMonster.setClientConnection(this);
	}

	/** Set the ClientAdapter that we will be interfacing with
	 * @param clientAdapter The object that we communicate with using the ClientAdapter interface
	 */
	public void setClientAdapter(ClientAdapter clientAdapter) {
		this.clientAdapter = clientAdapter;
	}

	/** Set the FIBSAttributes that we fetch our server attributes from
	 * @param fibsAttributes The FIBSAtrributes that we need.
	 */
	public void setFibsAttributes(FIBSAttributes fibsAttributes) {
		this.fibsAttributes = fibsAttributes;
	}

	/** Reset the state of the cookie monster
	 */
	public void resetFIBSCookieMonster() {
		this.cookieMonster.resetFIBSCookieMonster();
	}

	public void pushBack(String s) {
		pushbackString = s;
	}

	/** Queue a message up for transmission to the server
	 * @param s The message to send (must include eol).
	 */
	public	void sendMessage(String s) {
		if (s == null)
			throw new RuntimeException("Can't send a null message");
		if (DEBUG)
			System.out.println("User sent '" + s + "'");
		accessOutMessages(s);
	}

	/** Kill the connection to the server.
	 */
	public void shutDown() {
		shuttingDown = true;
		try {
			if (os != null)
				os.close();
			os = null;
			if (is != null)
				is.close();
			is = null;
			if (sock != null) {
				sock.shutdownInput();
				sock.shutdownOutput();
				sock.close();
			}
			sock = null;
		} catch (SocketException sex) {	
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/** Either append a message to the end, or remove a message from the head
	 * @param s a network message or null
	 */
	private void accessOutMessages(String s) {
		synchronized (listLock) {
			if (s == null)
				outMessages.remove();
			else
				outMessages.add(s);
		}
	}

	/* (non-Javadoc)
	 * @see java.lang.Thread#start()
	 */
	@Override
	public synchronized void run() {
		sock = null;
		String server = this.fibsAttributes.getServerName();
		int port = this.fibsAttributes.getServerPort();
		this.clientAdapter.writeSystemMessage(ClientAdapter.MessageRoute.SYSTEM, "Connecting to Server:" + eol);
		this.clientAdapter.writeSystemMessage(ClientAdapter.MessageRoute.SYSTEM, server + ":" + port + eol);
		try {
			sock = new Socket(server, port);
		} catch (UnknownHostException e) {
			this.clientAdapter.writeSystemMessage(ClientAdapter.MessageRoute.SYSTEM, "Can't connect: Unknown host");
			e.printStackTrace();
			return;
		} catch (IOException e) {
			this.clientAdapter.writeSystemMessage(ClientAdapter.MessageRoute.SYSTEM, "Can't connect: " + e.getLocalizedMessage());
			e.printStackTrace();
			return;
		}
		try {
			is = sock.getInputStream();
			os = sock.getOutputStream();
		
			while (!shuttingDown) {
				if (pushbackString != null) {
					String s = pushbackString;
					pushbackString = null;
					handleMessage(s);
					continue;
				}
				if (is.available() > 0) {
					readMessage();
					continue;
				}
				if (!outMessages.isEmpty()) {
					writeMessage();
					continue;
				}
				Thread.sleep(100);
			}
		} catch (IOException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		if (sock != null) {
			try {
				sock.shutdownInput();
				sock.shutdownOutput();
				sock.close();
			} catch (SocketException e) {
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	private	void readMessage() {
		byte[] b;
		try {
			int r = is.available();
			b = new byte[r];
			is.read(b);
		} catch (IOException e) {
			e.printStackTrace();
			shuttingDown = true;
			return;
		}
		String s = new String(b);
		if (DEBUG)
			System.out.println(s);
		if (DEBUG)
			this.clientAdapter.writeSystemMessageln(ClientAdapter.MessageRoute.DEBUG, s);
		boolean hasLeftOver = false;
		if (!s.endsWith("\r\n")) {
			if (DEBUG)
				System.out.println("Doesn't end with cr/lf" + eol);
			hasLeftOver = true;
		}
		String[] ss = s.split("\r\n");
		if (ss.length == 0)
			return;
		if (DEBUG)
			System.out.println("Got " + ss.length + " lines" + eol);
		if (leftover != null) {
			ss[0] = leftover + ss[0];
			leftover = null;
		}
		for (int i=0; i<ss.length; i++) {
			if (i == ss.length-1 && hasLeftOver && !ss[i].startsWith("login:"))
				leftover = ss[i];
			else
				handleMessage(ss[i]);
			while (pushbackString != null) {
				String t = pushbackString;
				pushbackString = null;
				handleMessage(t);
			}
		}
	}

	/** Send the login message to FIBS.
	 */
	public void sendLogin() {
		StringBuffer sb = new StringBuffer();
		sb.append("login ");
		sb.append(this.fibsAttributes.getAppSignature());
		sb.append(" 1008 ");
		sb.append(this.fibsAttributes.getUserName());
		sb.append(" ");
		sb.append(this.fibsAttributes.getUserPassword());
		sb.append(eol);
		this.sendMessage(sb.toString());
	}
 
	private	void writeMessage() {
		String s = outMessages.getFirst();
		if (this.fibsAttributes.isDisplayXmit()) {
			if (s.startsWith("login"))
				this.clientAdapter.writeSystemMessageln(ClientAdapter.MessageRoute.NETWORKOUT, "login...");
			else
				this.clientAdapter.writeSystemMessage(ClientAdapter.MessageRoute.NETWORKOUT, s);
		}
		if (this.fibsAttributes.isStdoutNetworkMessages())
			System.out.println("writeNetMessage: '" + s + "'");
		try {
			if (os == null) {
				shuttingDown = true;
				return;
			}
			os.write(s.getBytes());
		} catch (Exception e) {
			this.clientAdapter.writeSystemMessageln(ClientAdapter.MessageRoute.ERROR, "Disconnected from FIBS");
			this.clientAdapter.writeSystemMessageln(ClientAdapter.MessageRoute.ERROR, e.getLocalizedMessage());
			this.clientAdapter.connectionAborted();
			e.printStackTrace();
			shuttingDown = true;
		}
		this.accessOutMessages(null);
	}
	
	private	void handleMessage(String s) {
		if (s.length() > 1 && s.charAt(0) == 13)	// I have seen two '5's runon with just a cr between them,
			s = s.substring(1);						// not a crlf.  Check for that here.
		if (s.length() > 1 && s.charAt(0) == 10)	// I have seen two '5's runon with just a lf between them,
			s = s.substring(1);						// not a crlf.  Check for that here.
		int cookie = this.cookieMonster.fIBSCookie(s);
		if (this.fibsAttributes.isStdoutNetworkMessages())
			System.out.println("Handle: " + cookie + "'" + s + "'");
		if (this.pushbackString != null) {
			s = s.substring(0, s.length()-this.pushbackString.length());
			if (DEBUG)
				System.out.println("Truncating to '" + s + "'");
		}
		this.clientAdapter.dispatch(cookie, s);
	}
}
